import pkg_resources
import validate
import ConfigParser
from configobj import ConfigObj, flatten_errors
validator = validate.Validator()
import os.path

#TODO: check this works on Windows
ank_user_dir = os.path.join(os.path.expanduser("~"),  ".autonetkit")

def load_config():
    settings = ConfigParser.RawConfigParser()
    spec_file = pkg_resources.resource_filename(__name__,"/config/configspec.cfg")
    settings = ConfigObj(configspec=spec_file, encoding='UTF8')
# User's ANK settings
    user_config_file = os.path.join(ank_user_dir, "autonetkit.cfg")
    settings.merge(ConfigObj(user_config_file))
# ANK settings in current directory
    settings.merge(ConfigObj("autonetkit.cfg"))

    results = settings.validate(validator)
    if results != True:
        for (section_list, key, _) in flatten_errors(settings, results):
            if key is not None:
                print "Error loading configuration file:"
                print 'Invalid key "%s" in section "%s"' % (key, ', '.join(section_list))
                raise SystemExit
            else:
# ignore missing sections - use defaults
                #print 'The following section was missing:%s ' % ', '.join(section_list)
                pass
    return settings

#NOTE: this only gets loaded once package-wide if imported as import autonetkit.config
settings = load_config()



import json
import autonetkit.ank as ank_utils
import autonetkit.log as log
import autonetkit.ank_json
import netaddr

messaging = autonetkit.ank_messaging.AnkMessaging()

#TODO: allow slack in allocations: both for ASN (group level), and for collision domains to allow new nodes to be easily added

try:
    import cPickle as pickle
except ImportError:
    import pickle

def assign_asn_to_interasn_cds(G_ip):
    #TODO: rename to assign_asn_to_cds as also does intra-asn cds
    #TODO: make this a common function to ip4 and ip6
    G_phy = G_ip.overlay("phy")
    for collision_domain in G_ip.nodes("collision_domain"):
        neigh_asn = list(ank_utils.neigh_attr(G_ip, collision_domain, "asn", G_phy)) #asn of neighbors
        if len(set(neigh_asn)) == 1:
            asn = set(neigh_asn).pop() # asn of any neigh, as all same
        else:
            asn = ank_utils.most_frequent(neigh_asn) # allocate cd to asn with most neighbors in it
        collision_domain.asn = asn

    return

def allocate_ips(G_ip):
    log.info("Allocating Host loopback IPs")

# assign a /64 to each asn
    assign_asn_to_interasn_cds(G_ip)

    loopback_blocks = {}
    infra_blocks = {}

#TODO: check if need to do network address... possibly only for loopback_pool and infra_pool so maps to asn

    global_pool = netaddr.IPNetwork("2001:db8::/32").subnet(48)
    global_pool = netaddr.IPNetwork("::/32").subnet(64)
    global_pool.next() # network address
    [global_pool.next() for i in range(9)] # consume generator to start infra at "a", loopbacks at "b"
    loopback_pool = global_pool.next().subnet(80)
    infra_pool = global_pool.next().subnet(80)

    loopback_network = loopback_pool.next() # network address
    infra_network = infra_pool.next() # network address

    unique_asns = set(n.asn for n in G_ip)
    for asn in sorted(unique_asns):
        loopback_blocks[asn] = loopback_pool.next()
        infra_blocks[asn] = infra_pool.next()

    for asn, devices in G_ip.groupby("asn").items():
        subnets = infra_blocks[asn].subnet(96)
        subnets.next() # network address
        ptp_subnet = subnets.next().subnet(126)
        ptp_subnet.next() # network address
        all_cds = set(d for d in devices if d.collision_domain)
        ptp_cds = [cd for cd in all_cds if cd.degree() == 2]

        for cd in ptp_cds:
            subnet = ptp_subnet.next()
            hosts = subnet.iter_hosts()
            hosts.next() # drop .0 as a host address (valid but can be confusing)
            cd.subnet = subnet 
            for edge in cd.edges():
                edge.ip = hosts.next()

        non_ptp_cds = all_cds - set(ptp_cds)
        # break into /96 subnets
        for cd in non_ptp_cds:
            subnet = subnets.next()
            hosts = subnet.iter_hosts()
            hosts.next() # drop .0 as a host address (valid but can be confusing)
            cd.subnet = subnet 
            for edge in cd.edges():
                edge.ip = hosts.next()


        loopback_hosts = loopback_blocks[asn].iter_hosts()
        loopback_hosts.next() # drop .0 as a host address (valid but can be confusing)
        l3hosts = set(d for d in devices if d.is_l3device)
        for host in sorted(l3hosts):
            host.loopback = loopback_hosts.next()

    # Store allocations for routing advertisement
    G_ip.data.infra_blocks = infra_blocks
    G_ip.data.loopback_blocks = loopback_blocks

    jsontree = {}
    infra_tree = []
    loopback_tree = []

    for asn in unique_asns:
        children = []
        for cd in G_ip.nodes('collision_domain', asn = asn):
            cd_children = []
            for edge in cd.edges():
                cd_children.append( {
                    'name': "%s %s" % (edge.ip, edge.dst),
                    'subnet': edge.ip,
                    })
            children.append({
                    'name': "%s %s" % (cd.subnet, cd.id),
                    'subnet': cd.subnet,
                    'children': cd_children,
                    })
        asn_infra_tree = {'subnet': infra_blocks[asn], 'name': "AS%s" % asn, 'children': children}
        infra_tree.append( asn_infra_tree)


        children = []
        for host in sorted(G_ip.nodes("is_l3device", asn = asn)):
            children.append({
                'name': "%s %s" % (host.loopback, host.id),
                'subnet': host.loopback,
                })
        loopback_tree.append({
            'name': 'AS%s' % asn,
            'subnet': loopback_blocks[asn],
            'children': children,
            })

    infra_tree = {'name': "infra %s" % infra_network, 'children': infra_tree}
    loopback_tree = {'name': "loopback %s" % loopback_network, 'subnet': '1.2.3.4', 'children': loopback_tree}

    total_tree = {
            'name': "IPv6",
            'children': 
            [loopback_tree, infra_tree],
            #[infra_tree],
            }

    jsontree = json.dumps(total_tree, cls=autonetkit.ank_json.AnkEncoder, indent = 4)
    
    body = json.dumps({"ip_allocations": jsontree})
    #messaging.publish_compressed("www", "client", body)


#TODO: need to update with loopbacks if wish to advertise also - or subdivide blocks?

    #ip_tree.save()



import networkx as nx

def plot_dot(overlay_graph, edge_label_attribute = None, save = True, show = False):
    #TODO: turn off rescaling
    import subprocess
    graph = overlay_graph._graph.copy()
    graph_name = overlay_graph.name

    graph = graph.to_undirected()

    for node in overlay_graph:
        graph.node[node.node_id]['label'] = node.overlay.graphics.label
        (x, y) = node.overlay.graphics.x, node.overlay.graphics.y
        (x, y) = x/4, -1*y/4
        graph.node[node.node_id]['pos'] = "%s,%s" % (x, y)
        graph.node[node.node_id]['image'] = "icons/%s.png" % node.overlay.graphics.device_type
        graph.node[node.node_id]['shape'] = 'none'
        graph.node[node.node_id]['fontsize'] = 25
        graph.node[node.node_id]['labelloc'] = 'b' 
        graph.node[node.node_id]['fontcolor'] = 'white' 

    filename = "%s.dot" % graph_name
    nx.to_dot(graph, filename) 
    cmd = ["dot", "-Kfdp", filename, "-Tpdf", "-o", "%s.pdf" % graph_name]
    subprocess.call(cmd)

def plot_pylab(overlay_graph, edge_label_attribute = None, node_label_attribute = None, save = True, show = False):
    """ Plot a graph"""
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        print ("Matplotlib not found, not plotting using Matplotlib")
        return
    graph = overlay_graph._graph
    graph_name = overlay_graph.name

    try:
        import numpy
    except ImportError:
        print("Matplotlib plotting requires numpy for graph layout")
        return
    
    try:
        ids, x, y = zip(*[(node.id , node.overlay.graphics.x, node.overlay.graphics.y)
                for node in overlay_graph])
        x = numpy.asarray(x, dtype=float)
        y = numpy.asarray(y, dtype=float)
#TODO: combine these two operations together
        x -= x.min()
        x *= 1.0/x.max() 
        y -= y.min()
        y *= -1.0/y.max() # invert
        y += 1 # rescale from 0->1 not 1->0
#TODO: see if can use reshape-type commands here
        co_ords = zip(list(x), list(y))
        co_ords = [numpy.array([x, y]) for x, y in co_ords]
        pos = dict( zip(ids, co_ords))
    except KeyError:
        pos=nx.spring_layout(graph)

    plt.clf()
    cf = plt.gcf()
    ax=cf.add_axes((0,0,1,1))
    # Create axes to allow adding of text relative to map
    ax.set_axis_off() 
    font_color = "k"
    node_color = "#336699"
    edge_color = "#888888"

    nodes = nx.draw_networkx_nodes(graph, pos, 
                           node_size = 50, 
                           alpha = 0.8, linewidths = (0,0),
                           node_color = node_color,
                           cmap=plt.cm.jet)

    nx.draw_networkx_edges(graph, pos, arrows=False,
                           edge_color=edge_color,
                           alpha=0.8)
    
    if edge_label_attribute:
        edge_labels = {}
        for edge in overlay_graph.edges():
            attr = edge.get(edge_label_attribute)
            if attr:
                label = "%s" % attr
            else:
                label = ""

            edge_labels[(edge.src.node_id, edge.dst.node_id)] = label
        nx.draw_networkx_edge_labels(graph, pos, 
                            edge_labels = edge_labels,
                            font_size = 10,
                            font_color = font_color)

#TODO: where is anm from here? global? :/
    if node_label_attribute:
        labels = {}
        for n in overlay_graph:
            attr = n.get(node_label_attribute)
            if attr:
                label = "%s\n%s" % (n, attr)
            else:
                label = n

            labels[n.node_id] = label
    else:
        labels = dict( (n.node_id, n) for n in overlay_graph)

#TODO: need to strip out 
    nx.draw_networkx_labels(graph, pos, 
                            labels=labels,
                            font_size = 12,
                            font_color = font_color)

    title = "%s graph" % graph_name
    ax.text(0.02, 0.98, title, horizontalalignment='left',
                            weight='heavy', fontsize=16, color='k',
                            verticalalignment='top', 
                            transform=ax.transAxes)
    if show:
        plt.show()
    if save:
        filename = "%s.pdf" % graph_name
        plt.savefig(filename)

    plt.close()

